"""
A HTTP/2 interface to Kyoukai.

This uses https://python-hyper.org/projects/h2/en/stable/asyncio-example.html as a reference and a base. Massive thanks
to the authors of this page.

This server has some notable pitfalls:

    - It ignores any priority data that is sent by the client.
    - It is not paticularly fast (unbenchmarked, but it can be assumed to be slower than the httptools backend.)
    - It does not fully implement all events.

Additionally, this server is **untested** - it can and probably will fail horribly in production. Use with caution :)
"""
import asyncio
import collections
import logging
import ssl
import sys
import warnings
from functools import partial
from urllib.parse import urlsplit

import typing
from asphalt.core import Context
from h2.config import H2Configuration
from h2.connection import H2Connection
from h2.errors import ErrorCodes
from h2.events import (
    DataReceived, RequestReceived, WindowUpdated, StreamEnded
)
from h2.exceptions import ProtocolError
from werkzeug.datastructures import MultiDict
from werkzeug.wrappers import Request, Response

from kyoukai.asphalt import KyoukaiBaseComponent

# Sentinel value for the request being complete.
REQUEST_FINISHED = object()


def get_header(headers: typing.List[typing.Tuple[str, str]], name: str) -> str:
    """
    Gets a header from the list of headers, or None if it doesn't exist.
    """
    for header, value in headers:
        if header == name:
            return value


# WSGI helpers.
def create_wsgi_environment(r: 'H2State') -> MultiDict:
    """
    Creates a new WSGI environment from the RequestData provided.
    """
    # HTTP/2 special header path
    path = get_header(r.headers, ':path')

    # urlsplit the path
    sp_path = urlsplit(path)

    # HTTP/2 special header server name
    server_name = get_header(r.headers, ':authority')

    # try and split the port away
    try:
        server_name, port = server_name.split(':', 1)
    except ValueError as e:
        port = "8443"

    # HTTP/2 special header method
    method = get_header(r.headers, ":method")

    environ = MultiDict({
        # Basic items
        "PATH_INFO": sp_path.path,
        "QUERY_STRING": sp_path.query,
        "SERVER_PROTOCOL": "HTTP/2",
        "REQUEST_METHOD": method,
        # WSGI protocol things
        "wsgi.version": (1, 0),
        "wsgi.errors": sys.stderr,
        "wsgi.url_scheme": get_header(r.headers, ":scheme"),
        "wsgi.input": r,
        "wsgi.async": True,
        "wsgi.multithread": True,  # technically false sometimes, but oh well
        "wsgi.multiprocess": False,
        "wsgi.run_once": False,
        "SERVER_NAME": server_name,
        "SERVER_PORT": port,
        "REMOTE_ADDR": r._protocol.ip,
        "REMOTE_PORT": r._protocol.client_port
    })

    # Add the headers.
    for header, value in r.headers:
        if not header.startswith(":"):
            environ.add("HTTP_{}".format(header.replace("-", "_").upper()), value)

    return environ


class H2State:
    """
    A temporary class that is used to store request data for a HTTP/2 connection.

    This is also passed to the Werkzeug request to emit data.
    """

    def __init__(self, headers: list, stream_id, protocol: 'H2KyoukaiProtocol'):
        self.stream_id = stream_id
        self._protocol = protocol

        self.headers = headers

        # The queue of data.
        # This is a deque as reading from here is implicitly async.
        self.body = asyncio.Queue()

        # The data to emit.
        self._emit_headers = None
        self._emit_status = None

    def insert_data(self, data: bytes):
        """
        Writes data from the stream into the body.
        """
        self.body.put_nowait(data)

    async def read_async(self, to_end=True):
        """
        There's no good way to do this - WSGI isn't async, after all.

        However, you can use `read_async` on the Werkzeug request (which we subclass) to wait until the request has
        finished streaming.

        :param to_end: If ``to_end`` is specified, then read until the end of the request.
            Otherwise, it will read one data chunk.
        """
        data = b""
        if to_end:
            while True:
                d = await self.body.get()
                if d == REQUEST_FINISHED:
                    break
                data += d
        else:
            d = await self.body.get()
            if not d == REQUEST_FINISHED:
                data += d

        return data

    def read(self, size: int = -1) -> bytes:
        """
        Reads data from the request until it's all done.

        :param size: The maximum amount of data to receive.
        """
        # Thanks h2 docs page for the chunking inspiration
        # However, we're lazy and just shove it back on the front of the queue when we're done
        curr_data = b""
        while size < 0 or len(curr_data) < size:
            # size < 0 means read until we can't
            # b''.join is probably faster.
            try:
                curr_data += self.body.get_nowait()
            except asyncio.QueueEmpty:
                # no more data left to pop
                break

        # Get rid of any excess data.
        d = curr_data[:size]
        if len(curr_data) != len(d):
            # we have data left to read, so place it back on the left of the deque.
            self.body._queue.appendleft(curr_data[size:])

        return d

    def get_chunk(self) -> bytes:
        """
        Gets a chunk of data from the queue.
        """
        try:
            d = self.body.get_nowait()
            if d == REQUEST_FINISHED:
                return b""
        except asyncio.QueueEmpty:
            return b""

        return d

    def start_response(self, status: str, headers: typing.List[typing.Tuple[str, str]],
                       exc_info=None):
        """
        The ``start_response`` callable that is plugged into a Werkzeug response.
        """
        self._emit_status = status.split(" ")[0]
        self._emit_headers = headers

        # fake "write" callable
        return (lambda data: None)

    def get_response_headers(self):
        """
        Called by the protocol once the Response is writable to submit the request to the HTTP/2 state machine.
        """
        # This header must be added first!
        # All psuedo-headers come before the other headers.
        headers = [(":status", self._emit_status)]
        headers.extend(self._emit_headers)

        # Send the response headers.
        return headers

    def __iter__(self):
        return self

    def __next__(self):
        return self.get_chunk()


class H2KyoukaiComponent(KyoukaiBaseComponent):
    """
    A component subclass that creates H2KyoukaiProtocol instances.
    """
    def __init__(self, app, ssl_keyfile: str, ssl_certfile: str,
                 *, ip: str="127.0.0.1", port: int=4444):
        """
        Creates a new HTTP/2 SSL-based context.

        This will use the HTTP/2 protocol, disabling HTTP/1.1 support for this port. It is possible
        to run two ervers side-by-side, one HTTP/2 and one HTTP/1.1, if you run them on
        different ports.
        """
        super().__init__(app, ip, port)

        self.ssl_keyfile = ssl_keyfile
        self.ssl_certfile = ssl_certfile

        # override some needed config values
        self.cfg["ssl"] = True
        self.cfg["http2"] = True

    def get_protocol(self, ctx: Context, serv_info: tuple):
        return H2KyoukaiProtocol(self, ctx)

    async def start(self, ctx: Context):
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=self.ssl_certfile, keyfile=self.ssl_keyfile)
        ssl_context.set_alpn_protocols(["h2"])

        try:
            ssl_context.set_npn_protocols(["h2"])
        except NotImplementedError:
            # NPN protocol doesn't work here, so don't bother setting it
            pass

        protocol = partial(self.get_protocol, ctx, (self._server_name, self.port))
        self.app.finalize()
        self.server = await self.app.loop.create_server(protocol, self.ip, self.port,
                                                        ssl=ssl_context)
        self.logger.info("Kyoukai H2 serving on {}:{}".format(self.ip, self.port))


class H2KyoukaiProtocol(asyncio.Protocol):
    """
    The base protocol for Kyoukai, using H2.
    """

    def __init__(self, component, parent_context: Context):
        # The current component used by this connection.
        self.component = component

        self.parent_context = parent_context

        config = H2Configuration(client_side=False,
                                 header_encoding="utf-8")

        # The HTTP/2 state machine.
        self.conn = H2Connection(config=config)

        # The current transport for this connection.
        self.transport = None  # type: asyncio.WriteTransport

        # The current streams for this request.
        # This is a dictionary of stream_id -> data.
        self.streams = {}  # type: typing.Dict[H2State]

        # The current stream data queue.
        # This is a Queue of two-item tuples: (stream_data, stream_id).
        # These are plucked off of a Werkzeug request as they come in from Kyoukai.
        self.stream_data = collections.defaultdict(lambda *args: asyncio.Queue())

        # The current logger.
        self.logger = logging.getLogger("Kyoukai.HTTP2")

        # Client data.
        self.ip, self.client_port = None, None

        # If we're waiting to send data due to the window being exceeded.
        self._locked = collections.defaultdict(lambda *args: asyncio.Event())

        # The dictionary of stream tasks.
        self.stream_tasks = {}

    def raw_write(self, data: bytes):
        """
        Writes to the underlying transport.
        """
        try:
            return self.transport.write(data)
        except (OSError, ConnectionError):
            # We can't write to the transport, so we silently discard the error.
            # This probably means the client has disconnected.
            return None

    def connection_lost(self, exc):
        self.logger.debug("Connection lost from {}:{}".format(self.ip, self.client_port))

    def connection_made(self, transport: asyncio.WriteTransport):
        """
        Called when a connection is made.

        :param transport: The transport made by the connection.
        """
        # Set our own attributes, and update the HTTP/2 state machine.
        self.transport = transport
        try:
            self.ip, self.client_port = self.transport.get_extra_info("peername")
            self.logger.debug("Connection received from {}:{}".format(self.ip, self.client_port))
        except ValueError:
            # Sometimes socket.socket.getpeername() isn't available, so it tried to unpack a None.
            # Or, it returns None (wtf?)
            # So just provide some fake values.
            warnings.warn("getpeername() returned None, cannot provide transport information.")
            self.ip, self.client_port = None, None

        # Ensure that we are talking to a HTTP/2 client.
        # If we're not, they're gonna get really confused when we send
        # the HTTP/2 pre-amble, as they're expecting not that.

        ssl_sock = self.transport.get_extra_info("ssl_object")
        if ssl_sock is None:
            # We're not connecting over TLS.
            # For the sake of it, we're gonna assume that the client talks HTTP/2 instead of
            # HTTP/1.1,
            # or some other protocol.
            warnings.warn("HTTP/2 connection established over a non-TLS stream!")
        else:
            # Ensure we negotiated a `h2` connection.
            # This will check the ALPN protocol, but failing that, fall back to the NPN protocol.
            negotiated_protocol = ssl_sock.selected_alpn_protocol()
            if negotiated_protocol is None:
                negotiated_protocol = ssl_sock.selected_npn_protocol()

            if negotiated_protocol != "h2":
                # Close the connection - this isn't a HTTP/2 connection.
                # All we're gonna receive is junk data.
                # This uses error code 0x1 (PROTOCOL_ERROR) because it's technically a protocol
                # error.
                self.close(0x1)
                return

        # Send the HTTP2 preamble.
        self.logger.debug("Started the HTTP/2 connection.")
        self.conn.initiate_connection()
        self.raw_write(self.conn.data_to_send())

    def data_received(self, data: bytes):
        """
        Called when data is received from the underlying socket.
        """
        # Get a list of events by writing to the state machine.
        try:
            events = self.conn.receive_data(data)
        except ProtocolError:
            self.close(0x1)
            return
        # Find any data we need to send to the client first.
        self.transport.write(self.conn.data_to_send())

        # Then, switch upon the events we've received from the HTTP/2 client.
        for event in events:
            self.logger.debug("Received HTTP/2 event {.__class__.__name__}".format(event))
            # RequestReceived - headers have been received.
            # Handle it with ``request_received``.
            if isinstance(event, RequestReceived):
                self.request_received(event)
            # HTTP/2 body.
            # This will read data into the current stream.
            elif isinstance(event, DataReceived):
                self.receive_data(event)
            # The stream has ended.
            elif isinstance(event, StreamEnded):
                self.stream_complete(event)
            # The window for the event has updated.
            # This will unlock the event sender and continue sending data.
            elif isinstance(event, WindowUpdated):
                self.window_opened(event)

    def _processing_done(self, environ: dict, stream_id):
        """
        Callback for when processing is done on a request.
        """

        def _inner(fut: asyncio.Future):
            result = fut.result()  # type: Response

            # Get the H2State for this request.
            state = self.streams[stream_id]  # type: H2State

            # Get the app iterator.
            it = result(environ, state.start_response)
            headers = state.get_response_headers()

            # Send the headers.
            self.conn.send_headers(stream_id, headers, end_stream=False)

            # Place all the data from the app iterator on the queue.
            for i in it:
                self.stream_data[stream_id].put_nowait(i)

            # Add the sentinel value.
            self.stream_data[stream_id].put_nowait(REQUEST_FINISHED)

            # This will all be done with the sending task.

        return _inner

    async def sending_loop(self, stream_id):
        """
        This loop continues sending data to the client as it comes off of the queue.
        """
        while True:
            self._locked[stream_id].clear()
            data = await self.stream_data[stream_id].get()

            if data == REQUEST_FINISHED:
                # The request is finished - terminate the stream.
                self.conn.end_stream(stream_id)
                self.raw_write(self.conn.data_to_send())
                # This stream is dead, now.
                return

            # Buffer data - don't exceed the control flow window size.
            window_size = self.conn.local_flow_control_window(stream_id)
            chunk_size = min(window_size, len(data))
            data_to_send = data[:chunk_size]
            data_to_buffer = data[chunk_size:]

            if data_to_send:
                # Split it into chunks and send it out.
                max_size = self.conn.max_outbound_frame_size
                chunks = (
                    data_to_send[x:x + max_size]
                    for x in range(0, len(data_to_send), max_size)
                )
                for chunk in chunks:
                    self.conn.send_data(stream_id, chunk)
                self.raw_write(self.conn.data_to_send())

            if data_to_buffer:
                # Don't exceed flow window, set this data to be sent later.
                # Put it back on the left of the deque, then wait for our event to be set.
                self.stream_data[stream_id]._queue.appendleft(data)
                await self._locked[stream_id].wait()
                self._locked[stream_id].clear()
                continue

    # H2 callbacks
    def request_received(self, event: RequestReceived):
        """
        Called when a request has been received.
        """
        # Create the RequestData that stores this event.
        r = H2State(event.headers, event.stream_id, self)
        self.streams[event.stream_id] = r

        # Create the task that runs the app.
        app = self.component.app  # type: Kyoukai
        # Create the fake WSGI environment.
        env = create_wsgi_environment(r)
        request = app.request_class(environ=env)

        loop = app.loop
        t = loop.create_task(app.process_request(request, self.parent_context))
        self.stream_tasks[event.stream_id] = loop.create_task(self.sending_loop(event.stream_id))

        t.add_done_callback(self._processing_done(env, event.stream_id))

    def window_opened(self, event: WindowUpdated):
        """
        Called when a control flow window has opened again.
        """
        if event.stream_id:
            # Set the lock on the event, which will cause the sending_loop to wake up.
            self._locked[event.stream_id].set()
        else:
            # Unlock all events.
            for ev in self._locked.keys():
                ev.set()

    def receive_data(self, event: DataReceived):
        """
        Called when a request has data that has been received.
        """
        # Write into the RequestData for this event.
        try:
            req = self.streams[event.stream_id]
        except KeyError:
            # Reset the stream, because the client is stupid.
            self.conn.reset_stream(event.stream_id, ErrorCodes.PROTOCOL_ERROR)
        else:
            req.insert_data(event.data)

    def stream_complete(self, event: StreamEnded):
        """
        Called when a stream is complete.

        This will invoke Kyoukai, which will handle the request.
        """
        try:
            req = self.streams[event.stream_id]
        except KeyError:
            # shoo
            self.conn.reset_stream(event.stream_id, ErrorCodes.PROTOCOL_ERROR)
            return
        else:
            req.insert_data(REQUEST_FINISHED)

    def close(self, error_code: int=0):
        """
        Called to terminate the connection for some reason.

        This will close the underlying transport.
        """
        # Send a GOAWAY frame.
        self.conn.close_connection(error_code)
        self.raw_write(self.conn.data_to_send())

        self.transport.close()



"""
The core application.
"""

import asyncio
import logging
import traceback

from asphalt.core import Context, run_application
from werkzeug.exceptions import NotFound, MethodNotAllowed, HTTPException, InternalServerError, \
    BadRequestKeyError
from werkzeug.routing import RequestRedirect, Map
from werkzeug.wrappers import Request, Response

from kyoukai.asphalt import HTTPRequestContext
from kyoukai.blueprint import Blueprint

__version__ = "2.2.1.post1"
version_format = "Kyoukai/{}".format(__version__)

logger = logging.getLogger("Kyoukai")


class Kyoukai(object):
    """
    The Kyoukai type is the core of the Kyoukai framework, and the core of your web application 
    based upon the Kyoukai framework. It acts as a central router and request processor that 
    takes in requests from the protocols and returns responses.

    The application name is currently unused, but it is good practice to set it correctly anyway in 
    case it is used in future editions of Kyoukai.

    You normally create an application instance inside your component file, like so:

    .. code-block:: python

        from kyoukai.app import Kyoukai

        ... # setup code

        kyk = Kyoukai("my_app")
        kyk.register_blueprint(whatever)

        ... # other setup

        class MyContainer(ContainerComponent):
            async def start(self, ctx):
                self.add_component('kyoukai', KyoukaiComponent, ip="127.0.0.1", port=4444,
                                   app="app:app")

    Of course, you can also embed Kyoukai inside another app, by awaiting :meth:`Kyoukai.start`.
    """

    #: The class of request to spawn every request.
    #: This should be a subclass of :class:`werkzeug.wrappers.Request`.
    #: You can override this by passing ``request_class`` as a keyword argument to the app.
    request_class = Request

    #: The class of response to wrap automatically.
    #: This should be a subclass of :class:`werkzeug.wrappers.Response`.
    #: You can override this by passing ``response_class`` as a keyword argument to the app.
    response_class = Response

    def __init__(self,
                 application_name: str,
                 *,
                 server_name: str = None,
                 **kwargs):
        """
        :param application_name: The name of the application that is being created. This is \ 
            passed to the :class:`.Blueprint` being created as the root blueprint.  
            
            This is used in ``url_for``, for example, to get the endpoint of routes registered to \
            the root Blueprint.
            
        :param server_name: Keyword-only. The SERVER_NAME to use inside the fake WSGI environment \
            created for ``url_for``, if applicable.
            
        :param host_matching: Should host matching be enabled? This will be implicitly True if \
            ``host`` is not None.
            
        :param host: The host used for host matching, to be passed to the root Blueprint.
            By default, no host is used, so all hosts are matched on the root Blueprint.
            
        :param application_root: Keyword-only. The APPLICATION_ROOT to use inside the fake WSGI \ 
            environment created for ``url_for``, if applicable.
        
        :param loop: Keyword-only. The asyncio event loop to use for this app. If no loop is \ 
            specified it, will be automatically fetched using :meth:`asyncio.get_event_loop`.
        
        :param request_class: Keyword-only. The custom request class to instantiate requests with.
        :param response_class: Keyword-only. The custom response class to instantiate responses \ 
            with.

        :param context_class: Keyword-only. The :class:`.Context` subclass to use when creating a \
            context. Defaults to :class:`.HTTPRequestContext`.
        """
        self.name = application_name
        self.server_name = server_name

        # Try and get the loop from the keyword arguments - don't automatically perform
        # `get_event_loop`.
        self.loop = kwargs.pop("loop", None)
        if not self.loop:
            self.loop = asyncio.get_event_loop()

        # Create the root blueprint.
        self._root_bp = Blueprint(application_name, host=kwargs.get("host"),
                                  host_matching=kwargs.get("host_matching", False))

        # The current Component that is running this app.
        self.component = None

        # The Request/Response classes.
        self.request_class = kwargs.pop("request_class", self.request_class)
        self.response_class = kwargs.pop("response_class", self.response_class)

        #: The context class.
        self.context_class = kwargs.pop("context_class", HTTPRequestContext)

        # Is this app set to debug mode?
        self.debug = False

        # Any extra config.
        self.config = kwargs

    @property
    def root(self) -> Blueprint:
        """
        :return: The root Blueprint for the routing tree.
        """
        return self._root_bp

    def register_blueprint(self, child: Blueprint):
        """
        Registers a child blueprint to this app's root Blueprint.

        This will set up the Blueprint tree, as well as setting up the routing table when finalized.
        
        :param child: The child Blueprint to add. This must be an instance of :class:`~.Blueprint`.
        """
        return self.root.add_child(child)

    def finalize(self, **map_options) -> Map:
        """
        Finalizes the app and blueprints.

        This will calculate the current :class:`werkzeug.routing.Map` which is required for 
        routing to work.
        
        :param map_options: The options to pass to the Map for routing.
        """
        self.debug = self.config.get("debug", False)

        return self.root.finalize(**map_options)

    # Magic methods
    def __getattr__(self, item: str) -> object:
        """
        Override for __getattr__ to allow transparent mirroring onto the root Blueprint.

        For example, this allows doing ``@app.route`` instead of ``@app.root.route``.
        """
        if item in ("route", "errorhandler", "add_errorhandler", "add_route", "wrap_route",
                    "url_for", "before_request", "add_hook", "after_request",
                    "add_route_group"):
            return getattr(self.root, item)

        raise AttributeError("'{.__class__.__name__}' object has no attribute {}"
                             .format(self, item))

    def log_route(self, request: Request, code: int):
        """
        Logs a route invocation.

        :param request: The request produced.
        :param code: The response code of the route.
        """
        fmtted = "{} {} - {}".format(request.method, request.path, code)
        logger.info(fmtted)

    async def handle_httpexception(self, ctx: HTTPRequestContext, exception: HTTPException,
                                   environ: dict = None) -> Response:
        """
        Handle a HTTP Exception.

        :param ctx: The context of the request.
        :param exception: The HTTPException to handle.
        :param environ: The fake WSGI environment.
        
        :return: A :class:`werkzeug.wrappers.Response` that handles this response.
        """
        # Try and load the error handler recursively from the ctx.route.blueprint.
        bp = ctx.bp or self.root

        if environ is None:
            environ = ctx.environ

        cbl = lambda environ: Response("Internal server error during processing. Report this.",
                                       status=500)

        error_handler = bp.get_errorhandler(exception)
        if not error_handler:
            # Try the root Blueprint. This may happen if the blueprint requested isn't registered
            # properly in the root, for some reason.
            error_handler = self.root.get_errorhandler(exception)
            if not error_handler:
                # Just return the Exception's get_response.
                cbl = exception.get_response

        else:
            # Try and invoke the error handler to get the Response.
            # Wrap it in the try/except, so we can handle a default one.
            try:
                res = await error_handler.invoke(ctx, args=(exception,))
                # hacky way of unifying everything
                cbl = lambda environ: res
            except HTTPException as e:
                # why tho?
                logger.warning("Error handler function raised another error, using the "
                               "response from that...")
                cbl = e.get_response
            except Exception as e:
                logger.exception("Error in error handler!")
                cbl = InternalServerError(e).get_response
                # else:
                # result = wrap_response(result, self.response_class)

        try:
            result = cbl(environ=environ)
        except Exception:
            # ok
            logger.critical("Whilst handling a {}, response.get_response ({}) raised exception"
                            .format(exception.code, cbl), exc_info=True)
            result = Response("Critical server error. Your application is broken.",
                              status=500)

        if result.status_code != exception.code:
            logger.warning("Error handler {} returned code {} when exception was code {}..."
                           .format(error_handler.callable_repr, result.status_code,
                                   exception.code))

            return result

    async def process_request(self, request: Request, parent_context: Context) -> Response:
        """
        Processes a Request and returns a Response object.

        This is the main processing method of Kyoukai, and is meant to be used by one of the HTTP 
        server backends, and not by client code.

        :param request: \
            The :class:`werkzeug.wrappers.Request` object to process.
            A new :class:`~.HTTPRequestContext` will be provided to wrap this request inside of \ 
            to client code.

        :param parent_context: \
            The :class:`asphalt.core.Context` that is the parent context for this particular app. 
            It will be used as the parent for the HTTPRequestContext.

        :return: A :class:`werkzeug.wrappers.Response` object that can be written to the client \ 
            as a response.
        """
        if not self.root.finalized:
            raise RuntimeError("App was not finalized")

        # Create a new HTTPRequestContext.
        ctx = self.context_class(parent_context, request)
        ctx.app = self

        async with ctx:
            # Call match on our Blueprint to find the request.
            try:
                matched, params, rule = self.root.match(request.environ)
                ctx.params = params
                ctx.rule = rule
            except NotFound as e:
                # No route matched.
                self.log_route(ctx.request, 404)
                logger.debug("Could not resolve route for {request.path}."
                             .format(request=request))
                return await self.handle_httpexception(ctx, e, request.environ)
            except MethodNotAllowed as e:
                # 405 method not allowed
                self.log_route(ctx.request, 405)
                logger.debug("Could not resolve valid method for "
                             "{request.path} ({request.method})".format(request=request))
                return await self.handle_httpexception(ctx, e, request.environ)
            except RequestRedirect as e:
                # slashes etc
                # user code is not allowed to handle this
                self.log_route(ctx.request, 307)
                e.code = 307
                return e.get_response(request.environ)
            else:
                ctx.route_matched.dispatch(ctx=ctx)

            ctx.route = matched
            ctx.bp = ctx.route.bp

            result = None
            # Invoke the route.
            try:
                ctx.route_invoked.dispatch(ctx=ctx)
                # INTERCEPT
                if ctx.request.method.upper() == "OPTIONS":
                    # NO USER CODE HERE HEHEHEHEHE
                    # instead, we need to return an Allow: header
                    # kyoukai autocalcs this
                    result = Response(status=204)
                    result.headers["Allow"] = ",".join(x for x in ctx.rule.methods if x !=
                                                       "OPTIONS")
                else:
                    result = await matched.invoke(ctx, params=params)
            except BadRequestKeyError as e:
                logger.info("BadRequestKeyError: {}".format(' '.join(e.args)), exc_info=True)
                result = await self.handle_httpexception(ctx, e, request.environ)
            except HTTPException as e:
                fmtted = traceback.format_exception(type(e), e, e.__traceback__)
                logger.debug(''.join(fmtted))
                logger.info(
                    "Hit HTTPException ({}) inside function, delegating.".format(str(e))
                )
                result = await self.handle_httpexception(ctx, e, request.environ)
            except Exception as e:
                logger.exception("Unhandled exception in route function")
                new_e = InternalServerError()
                new_e.__cause__ = e
                result = await self.handle_httpexception(ctx, new_e, request.environ)
            else:
                ctx.route_completed.dispatch(ctx=ctx, result=result)
            finally:
                # result = wrap_response(result, self.response_class)
                if result:
                    # edge cases
                    self.log_route(ctx.request, result.status_code)

            # Update the Server header.
            result.headers["Server"] = version_format

            # list means wsgi response probably
            if not isinstance(result.response, (bytes, str, list)):
                result.set_data(str(result.response))

            result.headers["X-Powered-By"] = version_format

            # Return the new Response.
            return result

    async def start(self, ip: str = "127.0.0.1", port: int = 4444, *,
                    component=None, base_context: Context = None):
        """
        Runs the Kyoukai component asynchronously.

        This will bypass Asphalt's default runner, and allow you to run your app easily inside 
        something else, for example.

        :param ip: The IP of the built-in server.
        :param port: The port of the built-in server.
        :param component: The component to start the app with. This should be an instance of \
            :class:`~.KyoukaiComponent`.
                    
        :param base_context: The base context that the HTTPRequestContext should be started with.
        """
        if not base_context:
            base_context = Context()

        if not component:
            from kyoukai.asphalt import KyoukaiComponent
            self.component = KyoukaiComponent(self, ip, port)
        else:
            self.component = component

        # Start the app.
        await self.component.start(base_context)

    def run(self, ip: str = "127.0.0.1", port: int = 4444, *,
            component=None):
        """
        Runs the Kyoukai server from within your code.

        This is not normally invoked - instead Asphalt should invoke the Kyoukai component.
        However, this is here for convenience.
        """
        if not component:
            from kyoukai.asphalt import KyoukaiComponent
            component = KyoukaiComponent(self, ip, port)

        run_application(component)



"""
A blueprint is a container - a collection of routes.

Kyoukai uses Blueprints to create a routing tree - a tree of blueprints that are used to collect 
routes together and match routes easily.
"""
import logging

import typing
from werkzeug.exceptions import HTTPException
from werkzeug.routing import Map, Rule, Submount

from kyoukai.route import Route
from kyoukai.routegroup import RouteGroup, get_rg_bp


logger = logging.getLogger("Kyoukai")


class Blueprint(object):
    """
    A Blueprint is a "route container" - it contains 0 to N routes, and 0 to N child Blueprints 
    that inherit from the parent.
    """

    def __init__(self, name: str, parent: 'Blueprint' = None,
                 prefix: str = "", *,
                 host_matching: bool = False, host: str = None):
        """
        :param name: The name of this Blueprint.
            This is used when generating endpoints in the finalize stage.

        :param parent: The parent of this Blueprint.
            Parent blueprints will gather the routes of their children, and return a giant 
            :class:`werkzeug.routing.Map` object that contains all of the route maps in the children

        :param prefix: The prefix to be added to the start of every route name.
            This is inherited from parents - the parent prefix will also be added to the start of 
            every route.
            
        :param host_matching: Should host matching be enabled?
            This is implicitly True if ``host`` is non-None.
            
        :param host: The host of the Blueprint. Used for custom subdomain routing.
            If this is None, then this Blueprint will be used for all hosts.
        """
        #: The name of this Blueprint.
        self.name = name

        #: The parent :class:`~.Blueprint`.
        self._parent = parent

        #: Any children :class:`~.Blueprint` objects.
        self._children = []

        #: The current URL prefix.
        self._prefix = prefix

        #: If this Blueprint is finalized or not.
        #: Finalization of a blueprint means gathering all of the Maps, and compiling a routing
        #: table which stores the endpoints.
        self.finalized = False

        #: The list of routes.
        #: This is used in finalization.
        self.routes = []

        #: The :class:`~werkzeug.routing.Map` used for this blueprint.
        self.map = None  # type: Map

        #: The error handler dictionary.
        self.errorhandlers = {}

        #: The request hooks for this Blueprint.
        self._request_hooks = {}

        #: The host for this Blueprint.
        self._host = host
        self._host_matching = host_matching or self._host is not None

    @property
    def parent(self) -> "Blueprint":
        """
        :return: The parent Blueprint of this blueprint.
        """
        return self._parent

    @property
    def prefix(self) -> str:
        """
        :return: The prefix of this Blueprint.
        
        .. versionchanged:: 2.2.0
        
            Moved prefix combination to :attr:`.computed_prefix`.
        """
        return self._prefix

    @prefix.setter
    def prefix(self, value):
        self._prefix = value

    @property
    def computed_prefix(self) -> str:
        """
        :return: The combined prefix (parent + ours) of this Blueprint.
        
        .. versionadded:: 2.2.0
        """
        if self._parent:
            return self._parent.computed_prefix + self._prefix

        return self._prefix

    @property
    def tree_routes(self) -> 'typing.Generator[Route, None, None]':
        """
        :return: A generator that yields all routes from the tree, from parent to children.
        """
        for route in self.routes:
            yield route

        for child in self._children:
            yield from child.tree_routes

    @property
    def host(self) -> str:
        """
        :return: The host for this Blueprint, or the host of any parent Blueprint. 
        """
        if self._parent:
            return self._host or self.parent.host

        return self._host

    def get_submount(self) -> Submount:
        """
        Gets the :class:`werkzeug.routing.Submount` for this Blueprint.
        
        .. versionadded:: 2.2.0
        """
        inner = []
        # get child submounts
        for bp in self._children:
            inner.append(bp.get_submount())

        # get route submounts
        for route in self.routes:
            inner.append(route.get_submount())

        # create the submount
        sm = Submount(self._prefix,
                      rules=inner)

        return sm

    def traverse_tree(self) -> 'typing.Generator[Blueprint, None, None]':
        """
        Traverses the tree for children Blueprints.
        """
        for child in self._children:
            yield from child.traverse_tree()
            yield child

    def finalize(self, **map_options) -> Map:
        """
        Called on the root Blueprint when all Blueprints have been registered and the app is 
        starting.
        
        This will automatically build a :class:`werkzeug.routing.Map` of 
        :class:`werkzeug.routing.Rule` objects for each Blueprint.
        
        .. note::
        
            Calling this on sub-blueprints will have no effect, apart from generating a Map.  
            It is recommended to only call this on the root Blueprint.
            
        .. versionchanged:: 2.2.0
        
            This now uses submounts instead of a giant rule amalgamation.
        
        :param map_options: The options to pass to the created Map.
        :return: The :class:`werkzeug.routing.Map` created from the routing tree.
        """
        if self.finalized is True:
            return self.map

        submount = self.get_submount()
        logger.info("Scanned {} route(s) in the routing tree, building routing mapping."
                    .format(sum(1 for x in submount.get_rules(submount))))

        # Make a new Map() out of all of the routes.
        rule_map = Map([submount], host_matching=self._host_matching, **map_options)

        logger.info("Built route mapping with {} rules.".format(len(rule_map._rules)))

        # update self.map
        self.map = rule_map
        self.finalized = True

        return rule_map

    def add_child(self, blueprint: 'Blueprint') -> 'Blueprint':
        """
        Adds a Blueprint as a child of this one.
        This is automatically called when using another Blueprint as a parent.
        
        :param blueprint: The blueprint to add as a child.
        """
        self._children.append(blueprint)
        blueprint._parent = self
        return blueprint

    def route(self, routing_url: str, methods: typing.Sequence[str] = ("GET", "HEAD"),
              **kwargs):
        """
        Convenience decorator for adding a route.

        This is equivalent to:

        .. code-block:: python
        
            route = bp.wrap_route(func, **kwargs)
            bp.add_route(route, routing_url, methods)
           
        .. versionchanged:: 2.2.0
        
            Now accepts a Route as the function to decorate - this will add a new routing url and 
            method pair to :meth:`.Route.add_route`.
        """

        def _inner(func):
            if isinstance(func, Route):
                # don't re-wrap, only add the routing URL and methods
                func.add_path(routing_url, methods)
                return func

            route = self.wrap_route(func, **kwargs)
            self.add_route(route, routing_url, methods)
            return route

        return _inner

    def errorhandler(self, code: int, endcode: int = None, step: int = None):
        """
        Helper decorator for adding an error handler.

        This is equivalent to:

        .. code-block:: python

            route = bp.add_errorhandler(cbl, code)

        :param code: The error handler code to use.
        :param endcode: The end of the error code range to handle. Error handlers will be added \
            for all requests between code and endcode. If this is not provided, only one code \
            will be handled.

        :param step: The step for the error handler range.
        """

        def _inner(cbl):
            self.add_errorhandler(cbl, code, endcode, step)
            return cbl

        return _inner

    def wrap_route(self, cbl, *args, **kwargs) -> Route:
        """
        Wraps a callable in a Route.
        This is required for routes to be added.

        :param cbl: The callable to wrap.
        :return: A new :class:`~.Route` object.
        """
        rtt = Route(cbl, *args, **kwargs)
        return rtt

    def add_errorhandler(self, cbl, startcode: int, endcode: int = None, step: int = None):
        """
        Adds an error handler to the table of error handlers.

        A blueprint can only have one error handler per code. If it doesn't have an error handler 
        for that code, it will try to fetch recursively the parent's error handler.

        :param cbl: The callable error handler.
        :param startcode: The error code to handle, for example 404.
            This also represents the start of an error range, if endcode is not None.
        :param endcode: The end of the error code range to handle. Error handlers will be added \
            for all requests between startcode and endcode.
        :param step: The step for the error handler range.
        """
        # don't re-wrap it in a route if it's already a route
        rtt = cbl
        if not isinstance(rtt, Route):
            rtt = self.wrap_route(rtt, should_invoke_hooks=False)

        # if it's a single code, just use the one error handler
        if endcode is None:
            self.errorhandlers[startcode] = rtt
        else:
            # add a range of routes instead
            for i in range(startcode, endcode, step or 1):
                self.errorhandlers[i] = rtt

        rtt.bp = self
        return rtt

    def get_errorhandler(self, exc: typing.Union[HTTPException, int]) -> typing.Union[None, Route]:
        """
        Recursively acquires the error handler for the specified error.
        
        :param exc: The exception to get the error handler for.
            This can either be a HTTPException object, or an integer.

        :return: The :class:`~.Route` object that corresponds to the error handler, \ 
            or None if no error handler could be found.
        """
        if isinstance(exc, HTTPException):
            exc = exc.code

        try:
            return self.errorhandlers[exc]
        except KeyError:
            try:
                return self._parent.get_errorhandler(exc)
            except (KeyError, AttributeError):
                return None

    def get_hooks(self, type_: str):
        """
        Gets a list of hooks that match the current type.

        These are ordered from parent to child.

        :param type_: The type of hooks to get (currently "pre" or "post").
        :return: An iterable of hooks to run.
        """
        hooks = []
        if self._parent:
            hooks.extend(self._parent.get_hooks(type_))

        hooks.extend(self._request_hooks.get(type_, []))

        return hooks

    def add_hook(self, type_: str, hook):
        """
        Adds a hook to the current Blueprint.

        :param type_: The type of hook to add (currently "pre" or "post").
        :param hook: The callable function to add as a hook.
        """
        if type_ not in self._request_hooks:
            self._request_hooks[type_] = []

        self._request_hooks[type_].append(hook)
        return hook

    def after_request(self, func):
        """
        Convenience decorator to add a post-request hook.
        """
        return self.add_hook(type_="post", hook=func)

    def before_request(self, func):
        """
        Convenience decorator to add a pre-request hook.
        """
        return self.add_hook(type_="pre", hook=func)

    def add_route(self, route: Route, routing_url: str,
                  methods: typing.Sequence[str] = ("GET", "HEAD")):
        """
        Adds a route to the routing table and map.

        :param route: The route object to add.
        
            This can be gotten from :class:`.Blueprint.wrap_route`, or by 
            directly creating a Route object.

        :param routing_url: The Werkzeug-compatible routing URL to add this route under.
                    
            For more information, see http://werkzeug.pocoo.org/docs/0.11/routing/.

        :param methods: An iterable of valid method this route can be called with.

        :return: The unmodified :class:`~.Route` object.
        """
        # Add the routing path to the route
        route.add_path(routing_url, methods)
        # Add it to the list of routes to add later.
        self.routes.append(route)
        # Add the self to the route.
        route.bp = self

        return route

    def get_route(self, endpoint: str) -> 'typing.Union[Route, None]':
        """
        Gets the route associated with an endpoint.
        """
        for route in self.tree_routes:
            if route.get_endpoint_name() == endpoint:
                return route

        return None

    def add_route_group(self, group: 'RouteGroup'):
        """
        Adds a route group to the current Blueprint.
        
        :param group: The :class:`~.RouteGroup` to add. 
        """
        bp = get_rg_bp(group)
        self.add_child(bp)

        return self

    def url_for(self, environment: dict, endpoint: str, *,
                method: str = None, **kwargs) -> str:
        """
        Gets the URL for a specified endpoint using the arguments of the route.

        This works very similarly to Flask's ``url_for``.

        It is not recommended to invoke this method directly - instead, ``url_for`` is set on the 
        context object that is provided to your user function. This will allow you to invoke it 
        with the correct environment already set.

        :param environment: The WSGI environment to use to bind to the adapter.
        :param endpoint: The endpoint to try and retrieve.

        :param method: If set, the method to explicitly provide (for similar endpoints with \ 
            different allowed routes).

        :param kwargs: Keyword arguments to provide to the route.
        :return: The built URL for this endpoint.
        """
        bound = self.map.bind_to_environ(environment)

        # Build the URL from the endpoint.
        built_url = bound.build(endpoint, values=kwargs, method=method)

        return built_url

    def match(self, environment: dict) -> typing.Tuple[Route, typing.Container[typing.Any], Rule]:
        """
        Matches with the WSGI environment.
        
        .. warning::
            You should **not** be using this method yourself.
            
        .. versionchanged:: 2.2.0
            This will now return the :class:`werkeug.routing.Rule` as well.

        :param environment: The environment dict to perform matching with.
            You can use the ``environ`` argument of a Request to get the environment back.
            
        :return: A Route object, which can be invoked to return the right response, and the \
            parameters to invoke it with.
        """
        # Get the MapAdapter used for matching.
        adapter = self.map.bind_to_environ(environment)
        # Match the route, without catching any exceptions.
        # These exceptions are propagated into the app and handled there instead.
        rule, params = adapter.match(return_rule=True)

        route = self.get_route(rule.endpoint)

        return route, params, rule



"""
A high-performance HTTP/1.1 backend for the Kyoukai webserver using `httptools 
<https://github.com/MagicStack/httptools>`_.
"""
import asyncio
import base64
import gzip
import logging
import traceback
import warnings
import zlib
from io import BytesIO

import httptools
from asphalt.core import Context
from werkzeug.exceptions import MethodNotAllowed, BadRequest, InternalServerError
from werkzeug.wrappers import Response

from kyoukai.backends.http2 import H2KyoukaiProtocol
from kyoukai.wsgi import to_wsgi_environment, get_formatted_response

CRITICAL_ERROR_TEXT = """HTTP/1.0 500 INTERNAL SERVER ERROR
Server: Kyoukai
X-Powered-By: Kyoukai
X-HTTP-Backend: httptools
Content-Type: text/html; charset=utf-8
Content-Length: 310

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<title>Critical Server Error</title>
<h1>Critical Server Error</h1>
<p>An unrecoverable error has occurred within Kyoukai.
If you are the developer, please report this at <a href="https://github.com/SunDwarf/Kyoukai">the 
Kyoukai issue tracker.</a>
""".replace("\n", "\r\n")

HTTP_SWITCHING_PROTOCOLS = """HTTP/1.1 101 SWITCHING PROTOCOLS
Connection: Upgrade
Upgrade: h2c
Server: Kyoukai
X-Powered-By: Kyoukai
X-HTTP-Backend: httptools
Content-Length: 0

""".replace("\n", "\r\n")

HTTP_TOO_BIG = """HTTP/1.1 413 PAYLOAD TOO LARGE
Server: Kyoukai
X-Powered-By: Kyoukai
X-HTTP-Backend: httptools
Content-Length: 0

""".replace("\n", "\r\n")

HTTP_INVALID_COMPRESSION = """HTTP/1.1 400 BAD REQUEST
Server: Kyoukai
X-Powered-By: Kyoukai
X-HTTP-Backend: httptools
Content-Length: 25

Invalid compressed data
""".replace("\n", "\r\n")

PROTOCOL_CLASS = "KyoukaiProtocol"


class KyoukaiProtocol(asyncio.Protocol):  # pragma: no cover
    """
    The base protocol for Kyoukai using httptools for a HTTP/1.0 or HTTP/1.1 interface.
    """
    MAX_BODY_SIZE = 12 * 1024 * 1024

    def __init__(self, component, parent_context: Context,
                 server_ip: str, server_port: int):
        """
        :param component: The :class:`kyoukai.asphalt.KyoukaiComponent` associated with this request.
        :param parent_context: The parent context for this request.
            A new HTTPRequestContext will be derived from this.
        """

        self.component = component
        self.app = component.app
        self.parent_context = parent_context

        self.server_ip = server_ip
        self.server_port = server_port

        # Transport.
        # This is written to by our request when it's done.
        self.transport = None  # type: asyncio.WriteTransport

        # Request lock.
        # This ensures that requests are processed serially, and responded to in the correct order,
        # as the lock is released after processing a request completely.
        self.lock = asyncio.Lock()

        # The parser itself.
        # This is created per connection, and uses our own class.
        self.parser = httptools.HttpRequestParser(self)

        # A waiter that 'waits' on the event to clear.
        # Once the wait is over, it then delegates the request to the app.
        self.waiter = None  # type: asyncio.Task

        # The IP and port of the client.
        self.ip, self.client_port = None, None

        # Intermediary data storage.
        # This is a list because headers are appended as (Name, Value) pairs.
        # In HTTP/1.1, there can be multiple headers with the same name but different values.
        self.headers = []
        self.body = BytesIO()
        self.full_url = ""

        self.loop = self.app.loop
        self.logger = logging.getLogger("Kyoukai.HTTP11")

    def replace(self, other: type, *args, **kwargs) -> type:
        """
        Replaces our type with the other.
        """
        # Copy the properties we need.
        component = self.component
        app = component.app
        # Goodbye, ourselves!
        self.__class__ = other

        # Hello, not ourselves!
        # Call the new __init__.
        other.__init__(self, component, app, *args, **kwargs)

        return self

    # httptools callbacks
    def on_message_begin(self):
        """
        Called when a message begins.
        """
        self.body = BytesIO()
        self.headers = []
        self.full_url = ""

    def on_header(self, name: bytes, value: bytes):
        """
        Called when a header has been received.

        :param name: The name of the header.
        :param value: The value of the header.
        """
        self.headers.append((name.decode(), value.decode()))

    def on_headers_complete(self):
        """
        Called when the headers have been completely sent.
        """

    def on_body(self, body: bytes):
        """
        Called when part of the body has been received.

        :param body: The body text.
        """
        self.body.write(body)
        if self.body.tell() >= self.MAX_BODY_SIZE:
            # write a "too big" message
            self.write(HTTP_TOO_BIG)
            self.close()

    def on_url(self, url: bytes):
        """
        Called when a URL is received from the client.
        """
        self.full_url = url.decode("utf-8")

    def on_message_complete(self):
        """
        Called when a message is complete.
        This creates the worker task which will begin processing the request.
        """
        task = self.loop.create_task(self._wait_wrapper())
        self.waiter = task

    # asyncio procs
    def connection_made(self, transport: asyncio.WriteTransport):
        """
        Called when a connection is made via asyncio.

        :param transport: The transport this is using.
        """
        try:
            self.ip, self.client_port = transport.get_extra_info("peername")
            self.logger.debug("Connection received from {}:{}".format(self.ip, self.client_port))
        except ValueError:
            # Sometimes socket.socket.getpeername() isn't available, so it tried to unpack a None.
            # Or, it returns None (wtf?)
            # So just provide some fake values.
            warnings.warn("getpeername() returned None, cannot provide transport information.")
            self.ip, self.client_port = None, None

        self.transport = transport

        ssl_sock = self.transport.get_extra_info("ssl_object")
        if ssl_sock is not None:
            # Check if we negotiated a HTTP/2 connection.
            # This will check the ALPN protocol, but failing that, fall back to the NPN protocol.
            negotiated_protocol = ssl_sock.selected_alpn_protocol()
            if negotiated_protocol is None:
                negotiated_protocol = ssl_sock.selected_npn_protocol()

            if negotiated_protocol == "h2":
                # switch protocol to http/2 handler
                transport = self.transport
                new_self = self.replace(H2KyoukaiProtocol)  # type: H2KyoukaiProtocol
                # make sure the new connection
                type(new_self).connection_made(new_self, transport)
                return

        self.component.connection_made.dispatch(protocol=self)

    def connection_lost(self, exc):
        self.logger.debug("Connection lost from {}:{}".format(self.ip, self.client_port))
        self.component.connection_lost.dispatch(protocol=self)

    def data_received(self, data: bytes):
        """
        Called when data is received into the connection.
        """
        # Feed it into the parser, and handle any errors that might happen.
        try:
            self.parser.feed_data(data)
        except httptools.HttpParserInvalidMethodError as e:
            # Exceptions here are a bit tricky.
            # We can't simply call into the app to have it handle a 405/400 - there's no Request,
            # or environment. Instead, what we do is call a wrapper function
            # (handle_parser_exception) which will generate a fake WSGI environment, and then
            # automatically return a werkzeug httpexception that corresponds.
            self.handle_parser_exception(e)
        except httptools.HttpParserError as e:
            traceback.print_exc()
            self.handle_parser_exception(e)
        except httptools.HttpParserUpgrade as e:
            # It's a HTTP upgrade!
            # The only valid values of these that we wish to support (currently) are `h2c` and
            # `Websocket`. Currently, Kyoukai does not support websocket upgrade (soon™).
            # However, we cannot silently ignore websocket upgrades - we discard those for now and
            # disconnect.
            # Anything else, we also discard and disconnect.

            # httptools sucks, and only provides us an offset.
            # so what we do is hope the `Upgrade` header is in our header list.
            for name, header in self.headers:
                if name.lower() == "upgrade":
                    upgrade = header
                    break
            else:
                # thanks, we can't do shit.
                self.handle_parser_exception(e)
                return
            # If it's h2c, replace ourselves with the HTTP/2 client.
            if upgrade.lower() == "h2c":
                # Copy the transport into our local scope, as it becomes None after we've switched
                # type. Once we've replaced ourselves, call `connection_made` on the new type to
                # initialize.
                for name, header in self.headers:
                    if name.lower() == "http2-settings":
                        http2_settings = header
                        break
                else:
                    # can't find the http2_settings header, rip
                    self.handle_parser_exception(e)
                    return

                self.logger.info("Upgrading HTTP/1.1 to HTTP/2 connection.")

                # base64 decode the http2 settings packet
                decoded = base64.urlsafe_b64decode(http2_settings)
                # send a 101 switching protocols
                self.write(HTTP_SWITCHING_PROTOCOLS)

                transport = self.transport
                new_self = self.replace(H2KyoukaiProtocol)  # type: H2KyoukaiProtocol

                # update with the new settings
                new_self.conn.initiate_upgrade_connection(decoded)
                # this will have updated settings; now switch protocols.
                type(new_self).connection_made(new_self, transport)
                return

            # If it's Websocket, disconnect.
            if upgrade.lower() == "websocket":
                self.handle_parser_exception(e)
                return

            # If it's anything else, disconnect.
            self.handle_parser_exception(e)
            return

    # kyoukai handling
    def handle_parser_exception(self, exc: Exception):
        """
        Handles an exception when parsing.

        This will not call into the app (hence why it is a normal function, and not a coroutine).
        It will also close the connection when it's done.

        :param exc: The exception to handle.
        """
        if isinstance(exc, httptools.HttpParserInvalidMethodError):
            # 405 method not allowed
            r = MethodNotAllowed()
        elif isinstance(exc, httptools.HttpParserError):
            # 400 bad request
            r = BadRequest()
        elif isinstance(exc, httptools.HttpParserUpgrade):
            r = BadRequest(description="Invalid upgrade header.")
        else:
            # internal server error
            r = InternalServerError()

        # Make a fake environment.
        new_environ = to_wsgi_environment(headers=self.headers, method="", path="/",
                                          http_version="1.0", body=None)
        new_environ["SERVER_NAME"] = self.component.get_server_name()
        new_environ["SERVER_PORT"] = str(self.server_port)
        new_environ["REMOTE_ADDR"] = self.ip

        self.raw_write(get_formatted_response(r, new_environ))
        self.parser = httptools.HttpRequestParser(self)
        self.close()

    async def _wait_wrapper(self):
        try:
            if hasattr(self, "_wait"):
                await self._wait()
            else:
                return
        except:
            self.logger.critical("Error in Kyoukai's HTTP handling!", exc_info=True)
            self._raw_write(CRITICAL_ERROR_TEXT.encode())
            self.close()
        finally:
            # we might have change protocol by now.
            # if so, don't try and cancel the non-existant thing.
            if hasattr(self, "waiter"):
                self.waiter.cancel()
                self.waiter = None
                self.parser = httptools.HttpRequestParser(self)

    async def _wait(self):
        """
        The main core of the protocol.

        This constructs a new Werkzeug request from the headers.
        """
        # Check if the body has data in it by asking it to tell us what position it's seeked to.
        # If it's > 0, it has data, so we can use it. Otherwise, it doesn't, so it's useless.
        told = self.body.tell()
        if told:
            self.logger.debug("Read {} bytes of body data from the connection".format(told))
            self.body.seek(0)
            body = self.body
        else:
            body = None

        version = self.parser.get_http_version()
        method = self.parser.get_method().decode()

        for header, value in self.headers:
            # check if a content-encoding has been passed
            if header == "Content-Encoding" and body is not None:
                # no special encoding
                if value == "identity":
                    pass

                # gzip, decompress as such
                elif value == "gzip":
                    self.logger.debug("Decoding body data as gzip.")
                    try:
                        decompressed_data = gzip.decompress(body.read())
                    except zlib.error:
                        self.write(HTTP_INVALID_COMPRESSION)
                        self.close()
                        return

                    body = BytesIO(decompressed_data)

                # deflate, decompress as such
                elif value == "deflate":
                    z = zlib.decompressobj(6, zlib.DEFLATED, -zlib.MAX_WBITS, zlib.DEF_MEM_LEVEL, 0)
                    try:
                        decompressed_data = z.decompress(body.read())
                    except zlib.error:
                        self.write(HTTP_INVALID_COMPRESSION)
                        self.close()
                        return

                    body = BytesIO(decompressed_data)
                else:
                    self.logger.error("Unknown Content-Encoding sent by client: {}".format(value))

        new_environ = to_wsgi_environment(headers=self.headers, method=method, path=self.full_url,
                                          http_version=version, body=body)

        new_environ["kyoukai.protocol"] = self
        new_environ["SERVER_NAME"] = self.component.get_server_name()
        new_environ["SERVER_PORT"] = str(self.server_port)
        new_environ["REMOTE_ADDR"] = self.ip
        new_environ["REMOTE_PORT"] = self.client_port

        # Construct a Request object.
        new_r = self.app.request_class(new_environ, False)

        # Invoke the app.
        async with self.lock:
            try:
                result = await self.app.process_request(new_r, self.parent_context)
            except Exception:
                # not good!
                # write the scary exception text
                self.logger.exception("Error in Kyoukai request handling!")
                self._raw_write(CRITICAL_ERROR_TEXT.encode("utf-8"))
                return
            else:
                # Write the response.
                self.write_response(result, new_environ)
            finally:
                if not self.parser.should_keep_alive():
                    self.close()
                # unlock the event and remove the waiter
                self.parser = httptools.HttpRequestParser(self)

    # transport methods
    def close(self):
        return self.transport.close()

    def write_response(self, response: Response, fake_environ: dict):
        """
        Writes a Werkzeug response to the transport.
        """
        return self.raw_write(get_formatted_response(response, fake_environ))

    def write(self, data: str):
        """
        Writes data to the socket.
        """
        d = data.encode()
        return self.raw_write(d)

    def raw_write(self, data: bytes):
        """
        Writes data to the transport.
        """
        return self._raw_write(data)

    def _raw_write(self, data: bytes):
        """
        Does a raw write to the underlying transport, if we can.

        :param data: The data to write.
        """
        try:
            self.transport.write(data)
        except OSError:
            # connection might be closed...
            # just ignore it.
            return



"""
A uWSGI adapter for Kyoukai.

This allows you to run Kyoukai with uWSGI as the web server. It translates the WSGI protocol into 
Kyoukai itself.
"""

import asyncio
try:
    import greenlet
except ImportError:
    raise RuntimeError("uwsgi must be configured with the greenlet and asyncio workers enabled "
                       "for the uwsgi backend")

import functools

from asphalt.core import Context, ContainerComponent
from asphalt.core.component import component_types
from kyoukai.app import Kyoukai
from kyoukai.asphalt import KyoukaiBaseComponent
from werkzeug.wrappers import Request, Response


def uwsgi_entry_point(func):  # pragma: no cover
    """
    Wraps a function as a uWSGI entry point.

    This will automatically switch greenlets and set the result on a Future instance.
    :param func: The function to wrap.
    """

    @functools.wraps(func)
    async def _uwsgi_entry_point(self, current_greenlet: greenlet.greenlet, future: asyncio.Future,
                                 *args, **kwargs):
        """
        uWSGI wrapper entry point.
        """
        try:
            # Call the underlying wrapped function.
            result = await func(self, *args, **kwargs)
        except Exception as e:
            # Set the exception on the Future, allowing the web server to retrieve it.
            future.set_exception(e)
        else:
            # Set the result of the function on the Future.
            future.set_result(result)
        finally:
            # Switch back context.
            current_greenlet.switch()

    return _uwsgi_entry_point


class uWSGIAdapter(object):
    """
    The main adapter.

    To use uWSGI with Kyoukai, you must create an instance of this class with your app object, and 
    point uWSGI to the ``wsgi_application`` method.

    .. code:: python

        from kyoukai.backends.uwsgi import uWSGIAdapter
        from myapp.app import kyk

        adapter = uWSGIAdapter(kyk)

    This adapter also supports using Asphalt .yml files.

    .. code:: python

        adapter = uWSGIAdapter.from_component(component, "config.yml")

    Then, your app will be running under uWSGI.
    """

    def __init__(self, app: Kyoukai, base_context: Context = None):
        """
        :param app: The application this class is running.
        """
        self.app = app
        self.app.finalize()

        # The base_context is passed into the app's handle function to create the HTTPContext.
        # As such, we need to store it here.
        self.base_context = base_context or Context()

        # The event loop.
        # This is set in run_application if it is None, otherwise it is set in `from_component`,
        # to get the right event loop.
        # This is used on the futures.
        self.loop = None

    @classmethod
    def from_asphalt_config(cls, filename: str) -> 'uWSGIAdapter':
        """
        Produces a new uWSGIAdapter from an Asphalt config file.

        :param filename: The full path to the config file.
        """
        from ruamel import yaml
        with open(filename):
            config_data = yaml.load(filename)

        # Instantiate the root component
        try:
            component_config = config_data.pop('component')
        except KeyError:
            raise LookupError('missing configuration key: component') from None
        else:
            component = component_types.create_object(**component_config)  # type: ContainerComponent

        # Create a new Context.
        context = Context()

        # Run `start()` on the app.
        loop = asyncio.get_event_loop()
        # You best hope this doesn't block.
        loop.run_until_complete(component.start(context))

        # Find the KyoukaiComponent.
        for c in component.child_components:
            if isinstance(c, KyoukaiBaseComponent):
                break
        else:
            raise TypeError("Could not find KyoukaiComponent in component list")

        del component

        # Create a new adapter.
        klass = cls(c.app, context)
        return klass

    @uwsgi_entry_point
    async def enter_kyoukai(self, request: Request) -> Response:
        """
        The main entry point to enter the app.

        This will be running inside the app greenlet.

        :param request: The request object to handle.
        :return: A response object to return.
        """
        result = await self.app.process_request(request, self.base_context)

        return result

    def run_application(self, environment: dict, start_response: callable):
        """
        Main entry point for the application.
        Called upon every request.

        :param environment: The WSGI environment for the application.
        :param start_response: A callable which is used by the Response.
        :return: A Werkzeug Response object.
        """
        if self.loop is None:
            self.loop = asyncio.get_event_loop()

        # Produce an event from the environment.
        request = self.app.request_class(environment)

        # Create the future that is used for the result.
        # This is used so that WSGI can be implemented into asyncio.
        fut = asyncio.Future(loop=self.loop)

        # Get our current greenlet, which we pass into enter_kyoukai.
        g = greenlet.getcurrent()  # type: greenlet.greenlet

        # Make a task for the coroutine, and switch to the event loop context.
        self.loop.create_task(self.enter_kyoukai(g, fut, request))
        # bye-bye uwsgi! see you in a little while!
        g.parent.switch()

        # hello uwsgi!
        # we're back!

        # Get the result/exception from the future.
        # This will automatically re-raise. Kyoukai handles exceptions from user code, so this
        # *shouldn't* kill anything, only cause uWSGI to emit a 500.
        result = fut.result()  # type: Response

        # Write the response to the WSGI stream.
        return result(environment, start_response)



"""
Kyoukai is an async web framework for Python 3.5 and above.

.. currentmodule:: kyoukai

.. autosummary::
    :toctree: Kyoukai
        
    app
    backends
    asphalt
    blueprint
    route
    routegroup
    testing
    util
"""
import json

from werkzeug.exceptions import RequestEntityTooLarge
from werkzeug.formparser import FormDataParser


def _parse_json(parser: FormDataParser, stream, mimetype, content_length, options):
    if parser.max_content_length is not None and \
                    content_length is not None and \
                    content_length > parser.max_content_length:
        raise RequestEntityTooLarge()

    # json loads the stream and return it
    # todo: handle encoding
    data = stream.read().decode()
    return stream, json.loads(data), {}


FormDataParser.parse_functions["application/json"] = _parse_json

from kyoukai.app import Kyoukai, __version__
from kyoukai.asphalt import HTTPRequestContext, KyoukaiComponent
from kyoukai.blueprint import Blueprint
from kyoukai.route import Route
from kyoukai.routegroup import RouteGroup
from kyoukai.testing import TestKyoukai


__all__ = ("Kyoukai", "HTTPRequestContext", "KyoukaiComponent", "Blueprint", "Route", "RouteGroup",
           "TestKyoukai")



"""
Route groups are classes that allow you to group a set of routes together.

.. currentmodule:: kyoukai.routegroup
"""
import collections
import inspect
import typing


def get_rg_bp(group: 'RouteGroup'):
    """
    Gets the :class:`~.Blueprint` created from a :class:`~.RouteGroup`.
    """
    return getattr(group, "_{0.__name__}__blueprint".format(type(group)))


class RouteGroupType(type):
    """
    The metaclass for a route group.
    
    This is responsible for passing the keyword arguments to the metaclass.
    """

    def __new__(mcs, name, bases, class_body, **kwargs):
        """
        Override of `__new__` to ensure the __init__ signature is compatible.
        """
        return super().__new__(mcs, name, bases, class_body)

    def __init__(self, name, bases, class_body, **kwargs):
        """
        Override of `__init__` to store the blueprint params.
        """
        super().__init__(name, bases, class_body)
        self._bp_kwargs = kwargs

    def _init_blueprint(self, obb):
        """
        Initializes the Blueprint used by this route group.
        
        :param obb: The route group instance to intialize. 
        """
        # circular imports tm
        from kyoukai.blueprint import Blueprint
        bp = Blueprint(self.__name__, **self._bp_kwargs)
        # get all the method types that have a `.route` attr on them
        for name, value in inspect.getmembers(obb):
            # unwrap methods
            if not hasattr(value, "__func__"):
                continue

            func = value.__func__
            if getattr(func, "in_group", False) is True:
                # check the delegate type
                if func.rg_delegate == "route":
                    # wrap value, but use func attrs
                    # this preserves the method and `self`
                    rtt = bp.wrap_route(value, **func.route_kwargs)
                    rtt.routes = func.routes
                    rtt.bp = bp

                    # copy hooks
                    for type_, hooks in func.route_hooks.items():
                        for hook in hooks:
                            rtt.add_hook(type_, hook)

                    bp.routes.append(rtt)
                elif func.rg_delegate == "errorhandler":
                    # add the error handler using `errorhandler_code`
                    for code in func.errorhandler_codes:
                        bp.add_errorhandler(value, code)
                elif func.rg_delegate == "hook":
                    # add the hook
                    bp.add_hook(func.hook_type, value)

        setattr(obb, "_{.__name__}__blueprint".format(self), bp)

    def __call__(self, *args, **kwargs):
        obb = object.__new__(self)
        obb.__init__(*args, **kwargs)
        self._init_blueprint(obb)
        return obb


def route(url: str, methods: typing.Iterable[str] = ("GET", "HEAD"), **kwargs):
    """
    A companion function to the RouteGroup class. This follows :meth:`.Blueprint.route` in 
    terms of arguments, and marks a function as a route inside the class.
    
    This will return the original function, with some attributes attached:
    
        - ``in_group``: Marks the function as in the route group.
        - ``rg_delegate``: Internal. The type of function inside the group this is.
        - ``route_kwargs``: Keyword arguments to provide to ``wrap_route``.
        - ``route_url``: The routing URL to provide to ``add_route``.
        - ``route_methods``: The methods for the route.
        - ``route_hooks``: A defaultdict of route-specific hooks.
        
    Additionally, the following methods are added.
    
        - ``hook``: A decorator that adds a hook of type ``type_``.
        - ``before_request``: A decorator that adds a ``pre`` hook.
        - ``after_request``: A decorator that adds a ``post`` hook.
    
    .. versionadded:: 2.1.1
    
    .. versionchanged:: 2.1.3
    
        Added the ability to add route-specific hooks.
        
    .. versionchanged:: 2.2.0
        
        Now accepts an already edited function as the function to decorate - this will add a new \
        routing url and method pair to the :attr:`.Route.routes`.

    .. versionchanged:: 2.2.2

        Default methods changed to GET and HEAD.
    
    :param url: The routing URL of the route.
    :param methods: An iterable of methods for the route.
    """

    def inner(func):
        # add the required attrs which are used on a scan later
        func.in_group = True
        func.rg_delegate = "route"
        func.route_kwargs = kwargs

        # try and append to the routes
        # failing that, create a new list
        try:
            func.routes.append((url, methods))
        except AttributeError:
            func.routes = [(url, methods)]

        if not hasattr(func, "route_hooks"):
            func.route_hooks = collections.defaultdict(lambda: [])

            # helper for route-specific hooks.
            def hook(type_: str):
                def _inner2(hookfunc):
                    func.route_hooks[type_].append(hookfunc)
                    return hookfunc

                return _inner2

            func.hook = hook
            func.before_request = hook("pre")
            func.after_request = hook("post")

        return func

    return inner


def errorhandler(startcode: int, endcode: int = None, step: int = None):
    """
    A companion function to the RouteGroup class. This follows :meth:`.Blueprint.errorhandler` in 
    terms of arguments. 
    
    :param startcode: The error code to handle, for example 404.
        This also represents the start of an error range, if endcode is not None.
    :param endcode: The end of the error code range to handle. Error handlers will be added
        for all requests between startcode and endcode.
    :param step: The step for the error handler range.
    """

    def inner(func):
        func.in_group = True
        func.rg_delegate = "errorhandler"

        # less code here
        if endcode is None:
            codes = [startcode]
        else:
            codes = range(startcode, endcode, step or 1)

        for code in codes:
            try:
                func.errorhandler_codes.append(code)
            except AttributeError:
                func.errorhandler_codes = [code]

        return func

    return inner

# hooks


def hook(type_: str):
    """
    Marks a function as a hook.
    
    :param type_: The type of hook to mark. 
    """

    def inner(func):
        func.in_group = True
        func.rg_delegate = "hook"
        func.hook_type = type_

        return func

    return inner


def before_request(func):
    """
    Helper decorator to mark a function as a pre-request hook. 
    """
    return hook("pre")(func)


def after_request(func):
    """
    Helper decorator to mark a function as a post-request hook. 
    """
    return hook("post")(func)


class RouteGroup(object, metaclass=RouteGroupType):
    """
    A route group is a class that contains multiple methods that are decorated with the route 
    decorator. They produce a blueprint that can be added to the tree that includes all methods 
    in the route group.
     
    .. code-block:: python
    
        class MyGroup(RouteGroup, prefix="/api/v1"):
            def __init__(self, something: str):
                self.something = something
                
            @route("/ping")
            async def ping(self, ctx: HTTPRequestContext):
                return '{"response": self.something}'
                
    Blueprint parameters can be passed in the class call.
    
    To add the route group as a blueprint, use 
    :meth:`.Blueprint.add_route_group(MyGroup, *args, **kwargs)`.
    """



"""
Routes are wrapped function objects that are called upon a HTTP request.
"""
import collections
import inspect
import types
import typing

from werkzeug.exceptions import HTTPException
from werkzeug.routing import Rule, Submount
from werkzeug.wrappers import Response

from kyoukai.util import wrap_response


class Route(object):
    """
    A route object is a wrapped function.
    They invoke this function when invoked on routing and calling.
    """

    def __init__(self, function, *,
                 reverse_hooks: bool = False,
                 should_invoke_hooks: bool = True, do_argument_checking: bool = True,
                 endpoint: str = None):
        """        
        :param function: The underlying callable.
            This can be a function, or any other callable.

        :param reverse_hooks: If the request hooks should be reversed for this request (i.e child \
         to parent.)
        
        :param should_invoke_hooks: If request hooks should be invoked.
            This is automatically False for error handlers.
        
        :param do_argument_checking: If argument type and name checking is enabled for this route.
        
        :param endpoint: The custom endpoint for this route.
        """
        if not callable(function):
            raise TypeError("Route arg must be callable")

        #: The underlying callable for this route.
        self._callable = function

        #: If this route should do argument checking.
        self.do_argument_checking = do_argument_checking

        #: The :class:`~.Blueprint` this route is associated with.
        self.bp = None  # type: Blueprint

        #: A list of tuples (url, methods) for this Route.
        self.routes = []

        #: The custom endpoint for this route. Could be None.
        self.endpoint = endpoint

        self.reverse_hooks = reverse_hooks

        self.should_invoke_hooks = should_invoke_hooks

        #: Our own specific hooks.
        self.hooks = {}

    @property
    def callable_repr(self):
        return repr(self._callable)

    def add_path(self, url: str, methods: typing.Sequence[str] = ("GET", "HEAD")):
        """
        Adds a path to the current set of paths for this route.

        :param url: The routing URL to add.
        :param methods: An iterable of methods to use for this path.

        The URL and methods will be added as a pair.
        """
        self.routes.append((url, methods))
        return self

    def get_submount(self) -> Submount:
        """
        :return: A submount that represents this route.
        
        .. versionadded:: 2.2.0

        .. versionchanged:: 2.x.x

            Changed from getting a list of rules to a single submount object.
        """
        rules = []

        for url, methods in self.routes:
            # mutable list thx
            methods = list(methods)
            if "OPTIONS" not in methods:
                methods.append("OPTIONS")

            rule = Rule(url, methods=methods,
                        host=self.bp.host if self.bp is not None else None,
                        endpoint=self.get_endpoint_name())

            rules.append(rule)

        # pass an empty prefix since we don't care about adding that
        submount = Submount("", rules)
        return submount

    def get_endpoint_name(self, bp=None) -> str:
        """
        Gets the endpoint name for this route.
        
        :param bp: The :class:`.Blueprint` to use for name calculation.
        :return: The endpoint that can be used.
        """
        if self.endpoint is not None:
            return self.endpoint

        if bp is not None:
            prefix = bp.name
        else:
            prefix = self.bp.name if self.bp else ""

        return "{}.{}".format(prefix, self._callable.__name__)

    async def invoke_function(self, ctx, pre_hooks: list, post_hooks: list, params):
        """
        Invokes the underlying callable.
        This is for use in chaining routes.

        :param ctx: The :class:`~.HTTPRequestContext` to use for this route.
        :param pre_hooks: A list of hooks to call before the route is invoked.
        :param post_hooks: A list of hooks to call after the route is invoked.
        :param params: The parameters to pass to the function.
        :return: The result of the invoked function.
        """
        # Invoke the route function.
        try:
            # Invoke pre-request hooks, setting `ctx` equal to the new value.
            if self.should_invoke_hooks:
                for hook in pre_hooks:
                    _ = await hook(ctx)
                    if _ is not None:
                        ctx = _

            if isinstance(params, collections.Mapping):
                result = self._callable(ctx, **params)
            else:
                result = self._callable(ctx, *params)
            if inspect.isawaitable(result):
                result = await result

            result = wrap_response(result, ctx.app.response_class)
        except HTTPException as e:
            # This is a valid response type
            raise e
        else:
            # Invoke post-request hooks. These happen inside this `else` block because
            # post-request hooks are only meant to happen if the route invoked successfully.
            if self.should_invoke_hooks:
                for hook in post_hooks:
                    _ = await hook(ctx, result)
                    if _ is not None:
                        result = _

            return result

    def check_route_args(self, params: dict = None):
        """
        Checks the arguments for a route.

        :param params: The parameters passed in, as a dict.
        :raises TypeError: If the arguments passed in were not correct.
        """
        # Get the signature of our callable.
        sig = inspect.signature(self._callable, follow_wrapped=True)  # type: inspect.Signature
        # signature ignores the `self` param on methods, for some reason
        # not that i'm complaining
        f_nargs = len(sig.parameters) - 1
        if f_nargs < 0:
            raise TypeError("Route functions must take ctx argument")

        # If the lengths of the signature and the params are different, it's obviously wrong.
        if f_nargs != len(params):
            raise TypeError("Route takes {} args, passed in {} instead".format(f_nargs,
                                                                               len(params)))

        # Next, check that all the argument names in the signature are in the params,
        # so that they can be easily double star expanded into the function.
        for n, (name, arg) in enumerate(sig.parameters.items()):
            # Skip the first argument, because it is usually the HTTPRequestContext, and we don't
            # want to type check that.
            if n == 0:
                continue

            # prevent type checking `self` on methods
            if isinstance(self._callable, types.MethodType) and n == 1:
                continue

            assert isinstance(arg, inspect.Parameter)
            if arg.name not in params:
                raise ValueError("Argument {} not found in args for callable {}"
                                 .format(arg.name, self._callable.__name__))

            # Also, check that the type of the arg and the annotation matches.
            value = params[arg.name]
            if arg.annotation is None or arg.annotation is inspect.Parameter.empty:
                # No annotation, don't type check
                continue

            if not isinstance(value, arg.annotation):
                raise TypeError("Argument {} must be type {} (got type {})".format(
                    arg.name, arg.annotation, type(value))
                )

    def add_hook(self, type_: str, hook):
        """
        Adds a hook to the current Route.

        :param type_: The type of hook to add (currently "pre" or "post").
        :param hook: The callable function to add as a hook.
        """
        if type_ not in self.hooks:
            self.hooks[type_] = []

        self.hooks[type_].append(hook)

        return hook

    def get_hooks(self, type_: str):
        """
        Gets the hooks for the current Route for the type.
        
        :param type_: The type to get. 
        :return: A list of callables.
        """
        return self.hooks.get(type_, [])

    def before_request(self, func):
        """
        Convenience decorator to add a pre-request hook.
        """
        return self.add_hook(type_="pre", hook=func)

    def after_request(self, func):
        """
        Convenience decorator to add a post-request hook.
        """
        return self.add_hook(type_="post", hook=func)

    async def invoke(self, ctx, args: typing.Iterable[typing.Any] = (),
                     params: typing.Container = None) -> Response:
        """
        Invokes a route.
        This will run the underlying function.
        
        :param ctx: The :class:`~.HTTPRequestContext` which is used in this request.
        :param args: Any args to expand into the function.
        :param params: Any keyword params that are used in this request.

        :return: The result of the route's function.
        """
        if params is None:
            params = {}
        # Set some useful attributes on the ctx.
        ctx.route = self

        # Invoke the route function.
        if not params:
            params = {}

        if self.do_argument_checking and not args:
            self.check_route_args(params)
        else:
            params = list(args) + list(params.values())

        # Try and get the hooks.
        pre_hooks = self.bp.get_hooks("pre").copy()
        if self.reverse_hooks:
            pre_hooks = list(reversed(pre_hooks))

        post_hooks = self.bp.get_hooks("post").copy()
        if self.reverse_hooks:
            post_hooks = list(reversed(post_hooks))

        # merge pre and post hooks with the route-specific ones
        pre_hooks += self.get_hooks("pre").copy()
        post_hooks += self.get_hooks("post").copy()

        return await self.invoke_function(ctx, pre_hooks, post_hooks, params)



"""
A gunicorn (green unicorn) based WSGI wrapper for Kyoukai.

This is similar to the uWSGI wrapper in that it implements a WSGI application for gunicorn to attach 
to. This only works with the gaiohttp worker, as otherwise the worker will have to block whilst 
the event loop  runs.

The gaiohttp worker allows asyncio execution of a coroutine (in our case, Kyoukai.process_request), 
so the asyncio event loop should work like normal whilst allowing gunicorn to take the brunt of 
the application work.
"""
import asyncio

import typing
from asphalt.core import Context
from werkzeug.wrappers import Request, Response
from werkzeug.wsgi import ClosingIterator

try:
    import gunicorn
    import aiohttp
except ImportError:
    raise RuntimeError("gunicorn and aiohttp must be installed for the gunicorn backend")

from asphalt.core import ContainerComponent
from asphalt.core.component import component_types

from kyoukai.app import Kyoukai
from kyoukai.asphalt import KyoukaiBaseComponent


class GunicornAdapter(object):
    """
    This class should be instantiated to create a gunicorn-compatible WSGI class.

    The instance method ``run_application`` should be set as the WSGI application, which calls your 
    app.
    For example:
    
    .. code-block:: python

        kyk = Kyoukai("my_app")
        runner = GunicornAdapter(kyk)

        application = runner.run_application

    """

    @classmethod
    def from_asphalt_config(cls, filename: str) -> 'GunicornAdapter':
        """
        Creates a new :class:`GunicornAdapter` from an Asphalt component config file.
        """
        from ruamel import yaml
        with open(filename):
            config_data = yaml.load(filename)

        # Instantiate the root component
        try:
            component_config = config_data.pop('component')
        except KeyError:
            raise LookupError('missing configuration key: component') from None
        else:
            component = component_types.create_object(
                **component_config)  # type: ContainerComponent

        # Create a new Context.
        context = Context()

        # Run `start()` on the app.
        loop = asyncio.get_event_loop()
        # You best hope this doesn't block.
        loop.run_until_complete(component.start(context))

        # Find the KyoukaiComponent.
        for c in component.child_components:
            if isinstance(c, KyoukaiBaseComponent):
                break
        else:
            raise TypeError("Could not find KyoukaiComponent in component list")

        del component

        # Create a new adapter.
        klass = cls(c.app, context)
        return klass

    def __init__(self, app: Kyoukai, base_context: Context = None):
        """
        :param app: The current application to run.
        :param base_context: The base context which is used as the parent to the \ 
            HTTPRequestContext created.
        """
        self.app = app

        self._base_context = base_context
        if not self._base_context:
            # Create an empty context as the default base context.
            self._base_context = Context()

    async def _run_application(self, environ, start_response) -> ClosingIterator:
        """
        Main entry point of the application.

        This is wrapped into a :class:`asyncio.Task` by ``run_application`` which is awaited on to 
        run the app itself.
        """
        if not self.app.root.finalized:
            self.app.root.finalize()
        request = self.app.request_class(environ=environ)

        response = await self.app.process_request(request, self._base_context)
        # Return the Response object which will be iterated over by Gunicorn.

        c = response(environ, start_response)  # type: ClosingIterator
        return c

    def run_application(self, environment: dict, start_response: typing.Callable) -> asyncio.Task:
        """
        Runs Kyoukai for the current request.

        This is **not** a coroutine - it returns a single item (a Task), which is awaited on to get 
        the response.

        :param environment: The WSGI environment to run this request with.
        :param start_response: A callable that can be used to start the response.
        :return: A new :class:`asyncio.Task` that can be awaited on to get a response from the \
        application.
        """
        is_async = environment.get("wsgi.async", False)

        if not is_async:
            # Damnit. Return a WSGI response that ells the user they're stupid.
            r = Response("<h1>Error</h1><br/>You did not use the <code>gaiohttp</code> "
                         "gunicorn worker. This is an "
                         "error! Please switch to the gaiohttp worker instead.")
            r.headers["Content-Type"] = "text/html; charset=utf-8"
            r.status_code = 500
            return r(environment, start_response)

        coro = self._run_application(environment, start_response)

        loop = asyncio.get_event_loop()
        t = loop.create_task(coro)

        environment["kyoukai.task"] = t

        return t



"""
Misc utilities for usage inside the framework.
"""
import json
import typing

from werkzeug.wrappers import Response


# response utilities
def as_html(text: str, code: int = 200, headers: dict = None) -> Response:
    """
    Returns a HTML response.
    
    .. code-block:: python
    
        return as_html("<h1>Hel Na</h1>", code=403)
    
    :param text: The text to return. 
    :param code: The status code of the response.
    :param headers: Any optional headers.
    :return: A new :class:`werkzeug.wrappers.Response` representing the HTML. 
    """
    if headers is None:
        headers = {}

    r = Response(text, status=code, headers={"Content-Type": "text/html", **headers})
    return r


def as_plaintext(text: str, code: int = 200, headers: dict = None) -> Response:
    """
    Returns a plaintext response.
    
    .. code-block:: python
    
        return as_plaintext("hel yea", code=201)
    
    :param text: The text to return. 
    :param code: The status code of the response.
    :param headers: Any optional headers.
    :return: A new :class:`werkzeug.wrappers.Response` representing the text.
    """
    if headers is None:
        headers = {}

    r = Response(text, status=code, headers={"Content-Type": "text/plain", **headers})
    return r


def as_json(data: typing.Union[dict, list], code: int = 200, headers: dict = None,
            *, json_encoder: json.JSONEncoder = None, **kwargs) -> Response:
    """
    Returns a JSON response.
    
    .. code-block:: python
    
        return as_json({"response": "yes", "code": 201}, code=201)
    
    :param data: The data to encode.
    :param code: The status code of the response.
    :param headers: Any optional headers.
    :param json_encoder: The encoder class to use to encode.
    :return: A new :class:`werkzeug.wrappers.Response` representing the JSON.
    """
    if headers is None:
        headers = {}

    dumped = json.dumps(data, cls=json_encoder, **kwargs)
    r = Response(dumped, status=code, headers=headers)
    return r


def wrap_response(args, response_class: Response = Response) -> Response:
    """
    Wrap up a response, if applicable.
    This allows Flask-like `return "whatever"`.

    :param args: The arguments that are being wrapped.
    :param response_class: The Response class that is being used.
    """
    if not args:
        # Return a 204 NO CONTENT.
        return response_class("", status=204)

    if isinstance(args, tuple):
        # We enforce ``tuple`` here instead of any iterable.
        if len(args) == 1:
            # Only body, use 200 for the response code.
            return response_class(args[0], status=200)

        if len(args) == 2:
            # Body and status code.
            return response_class(args[0], status=args[1])

        if len(args) == 3:
            # Body, status code, and headers.
            return response_class(args[0], status=args[1], headers=args[2])

        raise TypeError("Cannot return more than 3 arguments from a view")

    if isinstance(args, response_class):
        # Return the bare response, unmodified.
        return args

    # Otherwise, wrap it in a response.
    return response_class(args)



"""
This file name is misleading. Kyoukai does NOT implement WSGI at all, except in the uWSGI wrapper. 
Instead, this file adds some utilities which convert from Werkzeug's WSGI magic to our normalized 
area.
"""
import sys
from io import BytesIO
from urllib.parse import urlsplit

import typing
from werkzeug.datastructures import MultiDict
from werkzeug.wrappers import Response


class SaneWSGIWrapper(object):
    """
    Forces a WSGI object to be in-line.
    """
    def __init__(self):
        self.headers = []
        self.real_body = b""

        self.status = None

    def start_response(self, status, headers: typing.List[tuple], exc_info=None):
        """
        Used as the ``start_response`` callable when interacting with WSGI devices.
        """
        self.status = status
        self.headers = headers

    def unfuck_iterable(self, i: typing.Iterable):
        """
        Unfucks the WSGI iterable into a single body.
        """
        for part in i:
            self.real_body += part

    def format(self):
        base = "HTTP/1.1 {status}\r\n{headers}\r\n"

        headers_fmt = ""
        # Calculate headers
        for name, val in self.headers:
            headers_fmt += "{}: {}\r\n".format(name, val)

        x = base.format(status=self.status, headers=headers_fmt).encode()
        x += self.real_body

        return x

    def __str__(self):
        """
        Converts this into the string format.
        """
        return self.format()


def to_wsgi_environment(headers: list, method: str, path: str,
                        http_version: str, body: BytesIO = None) -> MultiDict:
    """
    Produces a new WSGI environment from a set of data that is passed in.

    This will return a dictionary that is directly compatible with Werkzeug's Request wrapper.

    .. code-block:: python

        d = to_wsgi_environment([("Host", "127.0.0.1"), "GET", "/", None)
        request = werkzeug.wrappers.Request(d)

    :param headers: The headers of the HTTP request.
    :param method: The HTTP method of this request, e.g GET or POST.
    :param path: The HTTP path to get, in raw form.
        This should NOT be urldecoded, as the path is manually decoded.

    :param http_version: The HTTP version to use.
    :param body: A :class:`BytesIO` representing the body wrapper for this dict, or None if there \ 
        is no request body.

    :return: A new dict containing the fake WSGI environment.
    """
    if isinstance(headers, dict):
        headers = headers.items()

    # urlsplit the path
    sp_path = urlsplit(path)

    if body is None:
        body = BytesIO()

    environ = MultiDict({
        # Basic items
        "PATH_INFO": sp_path.path,
        "QUERY_STRING": sp_path.query,
        "SERVER_PROTOCOL": "HTTP/%s" % http_version,
        "REQUEST_METHOD": method,
        # WSGI protocol things
        "wsgi.errors": sys.stderr,
        "wsgi.url_scheme": "http",
        "wsgi.input": body,
        "wsgi.input_terminated": True,
        "wsgi.async": True,
        "wsgi.multithread": True,  # technically false sometimes, but oh well
        "wsgi.multiprocess": False,
        "wsgi.run_once": False
    })
    environ["wsgi.version"] = (1, 0)

    for header, value in headers:
        name = header.upper().replace("-", "_")
        if header not in ("Content-Type", "Content-Length"):
            name = "HTTP_{}".format(name)

        environ.add(name, value)

    return environ


def get_formatted_response(response: Response, environment: dict) -> bytes:
    """
    Transform a Werkzeug response into a HTTP response that can be sent back down the wire.

    :param response: The response object to transform.
    :return: Bytes of text that can be sent to a client.
    """
    wrapper = SaneWSGIWrapper()
    iterator = response(environment, wrapper.start_response)
    wrapper.unfuck_iterable(iterator)

    return wrapper.format()



"""
Asphalt wrappers for Kyoukai.
"""
import abc
import importlib
import logging
import socket
import ssl as py_ssl
from functools import partial

from asphalt.core import resolve_reference, Context
from asphalt.core.component import Component
from asphalt.core.event import Signal, Event
from werkzeug.routing import Rule
from werkzeug.wrappers import Request, Response

from kyoukai.blueprint import Blueprint
from kyoukai.route import Route


# Asphalt events.
class ConnectionMadeEvent(Event):  # pragma: no cover
    """
    Dispatched when a connection is made to the server.

    This does NOT fire when using WSGI workers.

    This has the protocol as the ``protocol`` attribute.
    """

    def __init__(self, source, topic, *, protocol):
        super().__init__(source, topic)
        self.protocol = protocol


class ConnectionLostEvent(ConnectionMadeEvent):  # pragma: no cover
    """
    Dispatched when a connection is lost from the server.

    This does NOT fire when using WSGI workers.

    This has the protocol as the ``protocol`` attribute.
    """


class CtxEvent(Event):  # pragma: no cover
    def __init__(self, source, topic, *, ctx: 'HTTPRequestContext'):
        super().__init__(source, topic)
        self.ctx = ctx


class RouteMatchedEvent(CtxEvent):  # pragma: no cover
    """
    Dispatched when a route is matched.

    This has the context as the ``ctx`` attribute, and the route can be accessed via ``ctx.route``.
    """


class RouteInvokedEvent(CtxEvent):  # pragma: no cover
    """
    Dispatched when a route is invoked.

    This has the context as the ``ctx`` attribute.
    """


class RouteReturnedEvent(CtxEvent):  # pragma: no cover
    """
    Dispatched after a route has returned.

    This has the context as the ``ctx`` attribute and the response as the ``result`` attribute.
    """

    def __init__(self, source, topic, *, ctx, result: Response):
        super().__init__(source, topic, ctx=ctx)
        self.result = result


class KyoukaiBaseComponent(Component, metaclass=abc.ABCMeta):  # pragma: no cover
    """
    The base class for any component used by Kyoukai.

    This one does not create a Server instance; it should be used when you are using a different 
    HTTP server backend.
    """
    connection_made = Signal(ConnectionMadeEvent)
    connection_lost = Signal(ConnectionLostEvent)

    def __init__(self, app, ip: str = "127.0.0.1", port: int = 4444, **cfg):
        from kyoukai.app import Kyoukai
        if not isinstance(app, Kyoukai):
            app = resolve_reference(app)

        #: The application object for a this component.
        self.app = app

        #: The IP address to boot the server on.
        self.ip = ip

        #: The port to boot the server on.
        self.port = port

        #: The config file to use.
        self.cfg = cfg

        #: The :class:`asyncio.Server` instance that is serving us today.
        self.server = None

        #: The base context for this server.
        self.base_context = None  # type: Context

        #: The backend to use for the HTTP server.
        self.backend = self.cfg.get("backend", "kyoukai.backends.httptools_")

        self.logger = logging.getLogger("Kyoukai")

        self._server_name = app.server_name or socket.getfqdn()

    @abc.abstractmethod
    async def start(self, ctx: Context):
        """
        Overridden in subclasses to spawn a new server.
        """

    def get_server_name(self):
        """
        :return: The server name of this app.
        """
        return self.app.server_name or self._server_name

    def get_protocol(self, ctx: Context, serv_info: tuple):
        """
        Gets the protocol to use for this webserver.
        """
        if not hasattr(self, "_cached_mod"):
            mod = importlib.import_module(self.backend)
            self._cached_mod = mod

        server = getattr(self._cached_mod, self._cached_mod.PROTOCOL_CLASS)
        proto = server(self, ctx, *serv_info)
        ctx.protocol = proto
        return proto


class KyoukaiComponent(KyoukaiBaseComponent):  # pragma: no cover
    """
    A component for Kyoukai.
    This includes the built-in HTTP server.  
    
    .. versionchanged:: 2.2
    
        Passing ``run_server`` as False will not run the inbuilt web server.
    """
    connection_made = Signal(ConnectionMadeEvent)
    connection_lost = Signal(ConnectionLostEvent)

    def __init__(self, app, ip: str = "127.0.0.1", port: int = 4444,
                 **cfg):
        """
        Creates a new component.

        :param app: The application object to use.
            This can either be the real application object, or a string that resolves to a \
            reference for the real application object.

        :param ip: If using the built-in HTTP server, the IP to bind to.
        :param port: If using the built-in HTTP server, the port to bind to.
        :param cfg: Additional configuration.
        """
        super().__init__(app, ip, port, **cfg)

        self.app.config.update(self.cfg)

        for key, value in cfg.items():
            setattr(self, key, value)

    def get_server_name(self):
        """
        :return: The server name of this app.
        """
        return self.app.server_name or self._server_name

    async def start(self, ctx: Context):
        """
        Starts the webserver if required.

        :param ctx: The base context.
        """
        self.base_context = ctx

        ssl_context = None

        if self.cfg.get("ssl", {}):
            ssl = self.cfg["ssl"]
            if ssl.get("enabled") is True:
                ssl_context = py_ssl.create_default_context(py_ssl.Purpose.CLIENT_AUTH)
                # override the ciphers
                ssl_context.set_ciphers(
                    "ECDH+CHACHA20:ECDH+CHACHA20:"         # CHACHA20 for newer openssl
                    "ECDH+AES128:RSA+AES128:"              # Standard AES
                    "ECDH+AES256:RSA+AES256:"              # Slower AES
                    "ECDH+3DES:RSA+3DES:"                  # 3DES for older systems
                    "!aNULL:!eNULL:!MD5:!DSS:!RC4")        # Disable insecure ciphers
                ssl_context.load_cert_chain(certfile=ssl["ssl_certfile"],
                                            keyfile=ssl["ssl_keyfile"])

                if self.cfg.get("http2", False) is True:
                    ssl_context.set_alpn_protocols(["h2"])

                    try:
                        ssl_context.set_npn_protocols(["h2"])
                    except NotImplementedError:
                        # NPN protocol doesn't work here, so don't bother setting it
                        pass

                self.logger.info("Using HTTP over TLS.")

        if self.cfg.get("run_server", True) is True:
            protocol = partial(self.get_protocol, ctx, (self._server_name, self.port))
            self.app.finalize()
            self.server = await self.app.loop.create_server(protocol, self.ip, self.port,
                                                            ssl=ssl_context)
            self.logger.info("Kyoukai serving on {}:{}.".format(self.ip, self.port))


class HTTPRequestContext(Context):
    """
    The context subclass passed to all requests within Kyoukai.
    """
    route_matched = Signal(RouteMatchedEvent)
    route_invoked = Signal(RouteInvokedEvent)
    route_completed = Signal(RouteReturnedEvent)

    def __init__(self, parent: Context, request: Request):
        super().__init__(parent)

        #: The :class:`~.Kyoukai` object this request is handling.
        self.app = None

        #: The :class:`werkzeug.wrappers.Request` object this request is handling.
        self.request = request

        #: The route parameters for this request. Usually contained by the routing URL.
        self.params = None  # type: dict

        #: The :class:`~.Route` object this request is for.
        self.route = None  # type: Route

        #: The :class:`~.Blueprint` object this request is for.
        self.bp = None  # type: Blueprint

        #: The :class:`werkzeug.routing.Rule` object associated with this request.
        self.rule = None  # type: Rule

        #: The WSGI environment for this request.
        self.environ = self.request.environ  # type: dict

        #: The :class:`asyncio.Protocol` protocol handling this connection.
        self.proto = None

    def url_for(self, endpoint: str, *, method: str = None, **kwargs):
        """
        A context-local version of ``url_for``.

        For more information, see the documentation on :meth:`~.Blueprint.url_for`.
        """
        return self.app.url_for(self.environ, endpoint, method=method, **kwargs)



import os
import sys

from setuptools import find_packages, setup

rootpath = os.path.abspath(os.path.dirname(__file__))


# Extract version
def extract_version(module='kyoukai'):
    version = None
    fname = os.path.join(rootpath, module, 'app.py')
    with open(fname) as f:
        for line in f:
            if line.startswith('__version__'):
                _, version = line.split('=')
                version = version.strip()[1:-1]  # Remove quotation characters.
                break
    return version


deps = [
    "httptools>=0.0.9,<0.1.0",
    "asphalt>=2.1.1,!=3.0.0",
    "werkzeug>=0.12.0,<0.13.0",
    "h2>=3.0.0,<3.1.0",
]

setup(
    name='Kyoukai',
    version=extract_version(),
    packages=[
        'kyoukai',
        'kyoukai.backends',

        # ext packages
        #'kyoukai.ext',
        #'kyoukai.ext.rest',
    ],
    url='https://mirai.veriny.tf',
    license='MIT',
    author='Laura Dickinson',
    author_email='l@veriny.tf',
    description='A fast, asynchronous web framework for Python 3.5+',
    #classifiers=[
    #    "Development Status :: 4 - Beta",
    #    "License :: OSI Approved :: MIT License",
    #    "Programming Language :: Python :: 3 :: Only",
    #    "Programming Language :: Python :: 3.5"
    #    "Programming Language :: Python :: 3.6",
    #    "Topic :: Internet",
    #    "Topic :: Software Development :: Libraries",
    #    "Topic :: Software Development :: Libraries :: Application Frameworks"
    #],
    install_requires=deps,
    test_requires=[
        "pytest",
        "pytest-asyncio"
    ],
    extras_require={
        "gunicorn": ["aiohttp>=2.2.0,<2.3.0", "gunicorn>=19.6.0"],
        "uwsgi": ["greenlet>=0.4.0"]
    }
)



"""
Testing helpers for Kyoukai.
"""
import random
from io import BytesIO

from asphalt.core import Context
from werkzeug.wrappers import Request, Response

from kyoukai.app import Kyoukai
from kyoukai.blueprint import Blueprint
from kyoukai.wsgi import to_wsgi_environment


class _TestingBpCtxManager(object):
    """
    A context manager that is returned from :meth:`~.TestKyoukai.testing_bp`. When entered, this will produce a new
    Blueprint object, that is then set onto the test application as the root blueprint.

    After exiting, it will automatically restore the old root Blueprint onto the application, allowing complete
    isolation of individual test routes away from eachother.
    """
    def __init__(self, app: 'TestKyoukai'):
        self.app = app

        self._old_root = self.app.root

    def __enter__(self) -> Blueprint:
        id = str(random.randint(0, 1000))
        name = "bp-{}".format(id)

        b = Blueprint(name)
        self.app._root_bp = b

        return b

    def __exit__(self, exc_type, exc_val, exc_tb):
        # reset the root blueprint
        del self.app._root_bp
        self.app._root_bp = self._old_root

        if exc_type:
            return False

        return True


class TestKyoukai(Kyoukai):
    """
    A special subclass that allows you to easily test your Kyoukai-based app.
    """

    def __init__(self, *args, base_context: Context = None, **kwargs):
        """
        :param base_context: The base context to use for all request testing.
        """
        super().__init__(*args, **kwargs)
        self.base_context = base_context

    @classmethod
    def wrap_existing_app(cls, other_app: Kyoukai, base_context: Context = None):
        """
        Wraps an existing app in a test frame.

        This allows easy usage of writing unit tests:

        .. code:: python

            # main.py
            kyk = Kyoukai("my_app")

            # test.py
            testing = TestKyoukai.wrap_existing_app(other_app)
            # use testing as you would normally

        :param other_app: The application object to wrap.
            Internally, this creates a new instance of ourselves, then sets the ``process_request`` 
            of the subclass to the copied object.

            This means whenever ``inject_request`` is called, it will use the old app's 
            process_request to run with,
            which will use the environment of the previous instance.

            Of course, if the old app has any side effects upon process_request, these side effects 
            will happen when the testing application runs as well, as the old app is completely 
            copied over.

        :param base_context: The base context to use for this.
        """
        new_object = cls("test_app", base_context=base_context)
        new_object.original_app = other_app
        new_object.process_request = other_app.process_request

        return new_object

    def testing_bp(self) -> _TestingBpCtxManager:
        """
        Context handler that allows ``with TestKyoukai.testing_bp() as bp:``

        You can then register items onto this new root blueprint until __exit__, which will then destroy the blueprint.
        """
        return _TestingBpCtxManager(self)

    async def inject_request(self, headers: dict, url: str, method: str = "GET", body: str = None) -> Response:
        """
        Injects a request into the test client.

        This will automatically create the correct context.

        :param headers: The headers to use.
        :param body: The body to use.
        :param url: The URL to use.
        :param method: The method to use.
        :return: The result.
        """
        if body is not None:
            body = BytesIO(body.encode())

        e = to_wsgi_environment(headers, method, url, http_version="1.1", body=body)
        e["SERVER_NAME"] = ""
        e["SERVER_PORT"] = ""

        r = Request(e)

        # for testing blueprints, etc
        # slow but it's a test so oh well
        self.finalize()

        return await self.process_request(r, self.base_context)
